﻿addon_query(_,_) :- assertz(bflow),model_to_uppaal,nl,print('#query_end#'),retractall(bflow).
bflow_header :- ( clause(bflow,true) -> print('addon:') ; true).
println(X) :- print(X),nl.

% For soundness-only checks in batch mode:
reduced_model_to_uppaal :- (re_yawl;true), model_to_uppaal.

% For all kind of analysis:
model_to_uppaal :- remove_events;
	(bflow_header,
	uppaal1,
	(uppaal2;true),
	((orjoin(_),uppaal3a);true),
	uppaal3b,
	((orjoin(_),uppaal3c);true),
	uppaal3d,
	uppaal4,
	uppaal5
	).




uppaal_startevent(X) :- startevent(X),arc(X,Out),
		print(X),
	print(' = StartNode('),
	print(X),print('_'),print(Out),
	print(');'),nl.

uppaal_endevent(X) :- endevent(X),arc(In,X),
	print(X),
	print(' = EndNode('),
	print(In),print('_'),print(X),
	print(');'),nl.

uppaal_activity(X) :- region(X),not(event(X)),
	arc(In,X),arc(X,Out),
	print(X),print(' = Activity(clock1,1,1,r0,'),
	print(In),print('_'),print(X),
	print(','),
	print(X),print('_'),print(Out),
	print(');'),nl.

uppaal_split(X) :- split(X),arc(In,X),
	type(X,Type),
	print(X),print(' = '),print(Type),print('split('),
	print(In),print('_'),print(X),
	print(','),
	findall(Out,(arc(X,Out),print(X),print('_'),print(Out),print(',')),Outlist),
	length(Outlist,OutlistLength),
	plus(OutlistLength,Dummies,7),
	(type(X,and),print_uppaal_dummies(Dummies,'alwayswrite');
	type(X,or),print_uppaal_dummies(Dummies,'nowrite');
	type(X,xor),print_uppaal_dummies(Dummies,'nowrite')),
	print(');'),nl.

uppaal_join(X) :- join(X),arc(X,Out),
	print(X),
	print(' = '),
	(
	(not(clause(all_joins_are_or_joins,true)),type(X,Type),print(Type))
	;
	(clause(all_joins_are_or_joins,true),print('or'))
	),
	print('join('),
	findall(In,(arc(In,X),print(In),print('_'),print(X),print(',')),Inlist),
	length(Inlist,InlistLength),
	plus(InlistLength,Dummies,7),
	print_uppaal_dummies(Dummies,'noread'),
	print(','),
	print(X),print('_'),print(Out),
	print(');'),nl.

print_uppaal_dummies(N,Dummytype) :- print(Dummytype),
	N > 1, !,print(','),
	N2 is N - 1,
	print_uppaal_dummies(N2,Dummytype).
print_uppaal_dummies(1,_).

uppaal_system_def :- print('system\t'),
	findall(X,(event(X),print(X),print(', ')),_),
	nl,print('\t'),
	findall(X,(region(X),not(event(X)),print(X),print(', ')),_), % Functions and Regions
	nl,print('\t'),
	findall(X,(split(X),print(X),print(', ')),_),
	nl,print('\t'),
	findall(X,(join(X),print(X),print(', ')),_),
	nl,print('\t'),
	print('Dummies;').


test_analyzability :- startevent(X), startevent(Y), X \== Y,
	  print('*** No analysis possible (model has more than one start event) ***').

test_analyzability :- split(X),findall(Y,arc(X,Y),L),length(L,Len), Len > 7,
	print('*** No analysis possible (model has split with more than 7 outgoing arcs) ***').

test_analyzability :- join(X),findall(Y,arc(Y,X),L),length(L,Len), Len > 7,
	print('*** No analysis possible (model has join with more than 7 inocming arcs) ***').

test_analyzability :- not(event(_)),
		print('*** No analysis possible (model is empty) ***').

uppaal1 :- print('<?xml version="1.0" encoding="utf-8"?><!DOCTYPE nta PUBLIC \'-//Uppaal Team//DTD Flat System 1.1//EN\' \'http://www.it.uu.se/research/group/darts/uppaal/flat-1_1.dtd\'><nta><declaration>').

uppaal2 :- test_analyzability.

uppaal2 :- remove_events;
	(print('bool split_at_work = false;'),nl,
	print('bool multiple_termination;'),nl,
	print('clock clock1;'),nl,
	print('int r0;'),nl,
	findall([X,Y],arc(X,Y),Arcs),length(Arcs,NumberOfArcs),
	print('const int number_of_arcs = '),print(NumberOfArcs),print(';'),nl,
	findall(X,function(X),Functions),length(Functions,NumberOfFunctions),
	print('const int number_of_activities = '),print(NumberOfFunctions),print(';'),nl,
	(
	(not(clause(all_joins_are_or_joins,true)),
	findall([Y,X],(orjoin(X),arc(Y,X)),ArcsToOrjoins),length(ArcsToOrjoins,NumberOfArcsToOrjoins),
	findall([Y,X],(arc(Y,X),not(orjoin(X))),ArcsToLocalElements))
	;
	(clause(all_joins_are_or_joins,true),
	findall([Y,X],(join(X),arc(Y,X)),ArcsToOrjoins),length(ArcsToOrjoins,NumberOfArcsToOrjoins),
	findall([Y,X],(arc(Y,X),not(join(X))),ArcsToLocalElements))
	),
	print('const int number_of_arcs_to_an_or_join = '),print(NumberOfArcsToOrjoins),print(';'),nl,
	print('urgent chan c[number_of_arcs+4];'),nl,
	print('const int noread = number_of_arcs;'),nl,
	print('const int nowrite = number_of_arcs+1;'),nl,
	print('const int alwaysread = number_of_arcs+2;'),nl,
	print('const int alwayswrite = number_of_arcs+3;'),nl,
	channels_for_arcs_to_orjoins(ArcsToOrjoins),
	channels_for_arcs_to_local_elements(ArcsToLocalElements,NumberOfArcsToOrjoins),
	print('bool t[number_of_arcs+4];'),nl,
	print('bool consumed[number_of_arcs+4];'),nl,
	print('bool no_joins_working()'),nl,
	print('{'),nl,
	print('   return forall (i : int[0,number_of_arcs]) !consumed[i];'),nl,
	print('}'),nl,
	print('// Will be set to 1 if an activity has been performed.'),nl,
	print('bool visited[number_of_arcs];'),nl,
	print('// Will be set to yes as soon as the execution has been started,'),nl,
	print('// i.e. a start event has forwarded a token:'),nl,
	print('bool started = false;'),nl,nl,
	print('// Returns yes if the execution has been completed, i.e. it has started and'),nl,
	print('// (after some execution) only end events have a token, i.e. there is no token'),nl,
	print('// somewhere else:'),nl,
	print('bool completed()'),nl,
	print('{'),nl,
	print('return started'),channels_for_arcs_that_does_not_lead_to_endevent,
	print(';'),nl,print('}'),
	(orjoin(_);(clause(all_joins_are_or_joins,true),join(_))),	% without an OR-join in the model, uppaal2 ends here.
	print('const bool s[number_of_arcs_to_an_or_join][number_of_arcs]'),nl,
	print('= {'),nl,
	print_uppaal_structure(ArcsToOrjoins,ArcsToLocalElements),
	print('};')).


channels_for_arcs_that_does_not_lead_to_endevent :-
findall([Y,X],(arc(Y,X),not(endevent(X)),print_t_component([Y,X])),_).

print_t_component([Y,X]) :- print(' &amp;&amp; !t['),print(Y),print('_'),print(X),print(']').



uppaal3a :- println('bool canfire(int in1, int in2, int in3, int in4, int in5, int in6, int in7, int out){
const int in[7] = {in1, in2, in3, in4, in5, in6, in7}; '),
	 println('// s_0 will be built such that it is true for the arcs with a token '),
	 println('// that have a path to an incoming arc of the OR-join which does not have a token
bool s_0[number_of_arcs]; '),
	 println('// s_1 will be built such that it is true for the arcs with a token '),
	 println('// that have a path to an incoming arc of the OR-join which have a token
bool s_1[number_of_arcs]; '),
	 println('// We search for tokens for which we have to wait (those who can reach an empty '),
	 println('// arc, but no arc with a token) '),
	 nl,
	 println('// Analyze all incoming arcs'),
	 println('// If an arc has a token: add all tokens that can reach this arc to s_1'),
	 println('// otherwise: add all tokens that can reach this arc to s_0'),
	 println('for(j: int[0,number_of_arcs-1]) {'),
	 println('	for(i: int[0,6]) {'),
	 println('		if (in[i] != noread) {'),
	 println('		if (t[in[i]]) 	{ s_1[j]:= t[j] &amp;&amp; (s_1[j] || s[in[i]][j]); }'),
	 println('		else		{ s_0[j]:= t[j] &amp;&amp; (s_0[j] || s[in[i]][j] ); }'),
	 println('}'),
	 println('}'),
	 println('// We have to wait for a token on arc j iff'),
	 println('// arc j can send a token to an ampty slot of the OR-join (i.e. s_0[j] == true) '),
	 println('// and arc j cannot send a token to a non-empty slot of the OR-join (i.e. s_1[j] == false) '),
	 println('s_0[j]:= s_0[j] &amp;&amp; !s_1[j];'),
	 println('}'),
	 nl,
	 println('// If the array s_0 has now an element that is true, this means that the OR-'),
	 println('// join has to wait for some token to arrive'),
	 println('return !exists(i: int[0,number_of_arcs-1]) s_0[i];'),
	 println('}').


uppaal3b :- println('</declaration>'),
	println('<template><name x="5" y="5">Activity</name><parameter>clock &amp;processclock, const int mintime, const int maxtime, int &amp;ressource, int in, int out</parameter><location id="id0" x="432" y="240"></location><location id="id1" x="216" y="240"><name x="184" y="208">working</name><label kind="invariant" x="144" y="256">processclock &lt;= maxtime</label></location><location id="id2" x="0" y="240"></location><init ref="id2"/><transition><source ref="id0"/><target ref="id2"/><label kind="synchronisation" x="264" y="296">c[out]!</label><label kind="assignment" x="360" y="296">t[out]:=true,t[in]:=false,
visited[in]=true</label><nail x="432" y="296"/><nail x="0" y="296"/></transition><transition><source ref="id1"/><target ref="id0"/><label kind="guard" x="256" y="200">processclock &gt;= mintime</label><label kind="assignment" x="288" y="216">ressource--</label></transition><transition><source ref="id2"/><target ref="id1"/><label kind="synchronisation" x="64" y="216">c[in]?</label><label kind="assignment" x="36" y="240">processclock = 0,ressource++</label></transition></template>'),
	println('<template><name x="5" y="5">StartNode</name><parameter>int out</parameter><location id="id3" x="240" y="272"></location><location id="id4" x="408" y="272"></location><init ref="id3"/><transition><source ref="id3"/><target ref="id4"/><label kind="synchronisation" x="296" y="248">c[out]!</label><label kind="assignment" x="256" y="280">t[out]=true,started=true</label></transition></template>'),
	println('<template><name x="5" y="5">EndNode</name><parameter>int in</parameter><location id="id5" x="256" y="240"><urgent/></location><location id="id6" x="528" y="240"><name x="488" y="200">mult_term</name></location><location id="id7" x="184" y="240"></location><location id="id8" x="392" y="240"><name x="374" y="202">finished</name></location><init ref="id7"/><transition><source ref="id5"/><target ref="id8"/><label kind="guard" x="280" y="216">!split_at_work</label><label kind="assignment" x="288" y="240">t[in]=false</label></transition><transition><source ref="id7"/><target ref="id5"/><label kind="synchronisation" x="200" y="216">c[in]?</label></transition><transition><source ref="id8"/><target ref="id6"/><label kind="synchronisation" x="432" y="216">c[in]?</label><label kind="assignment" x="368" y="248">multiple_termination = true</label></transition><transition><source ref="id8"/><target ref="id8"/><label kind="guard" x="334" y="186">no_joins_working()</label><nail x="390" y="218"/><nail x="422" y="23"/></transition></template>'),
	println('<template><name x="5" y="5">xorsplit</name><parameter>int in, int out1, int out2, int out3, int out4, int out5, int out6, int out7</parameter><location id="id9" x="192" y="240"></location><location id="id10" x="128" y="240"></location><init ref="id10"/><transition><source ref="id9"/><target ref="id10"/><label kind="guard" x="416" y="48">!split_at_work</label><label kind="synchronisation" x="136" y="48">c[out7]!</label><label kind="assignment" x="200" y="48">t[out7]=true,t[in]=false</label><nail x="192" y="72"/><nail x="128" y="72"/></transition><transition><source ref="id10"/><target ref="id9"/><label kind="synchronisation" x="144" y="216">c[in]?</label></transition><transition><source ref="id9"/><target ref="id10"/><label kind="guard" x="416" y="72">!split_at_work</label><label kind="synchronisation" x="136" y="72">c[out6]!</label><label kind="assignment" x="200" y="72">t[out6]=true,t[in]=false</label><nail x="192" y="96"/><nail x="128" y="96"/></transition><transition><source ref="id9"/><target ref="id10"/><label kind="guard" x="416" y="96">!split_at_work</label><label kind="synchronisation" x="136" y="96">c[out5]!</label><label kind="assignment" x="200" y="96">t[out5]=true,t[in]=false</label><nail x="192" y="120"/><nail x="128" y="120"/></transition><transition><source ref="id9"/><target ref="id10"/><label kind="guard" x="416" y="120">!split_at_work</label><label kind="synchronisation" x="136" y="120">c[out4]!</label><label kind="assignment" x="200" y="120">t[out4]=true,t[in]=false</label><nail x="192" y="144"/><nail x="128" y="144"/></transition><transition><source ref="id9"/><target ref="id10"/><label kind="guard" x="416" y="144">!split_at_work</label><label kind="synchronisation" x="136" y="144">c[out3]!</label><label kind="assignment" x="200" y="144">t[out3]=true,t[in]=false</label><nail x="192" y="168"/><nail x="128" y="168"/></transition><transition><source ref="id9"/><target ref="id10"/><label kind="guard" x="416" y="168">!split_at_work</label><label kind="synchronisation" x="136" y="168">c[out2]!</label><label kind="assignment" x="200" y="168">t[out2]=true,t[in]=false</label><nail x="192" y="192"/><nail x="128" y="192"/></transition><transition><source ref="id9"/><target ref="id10"/><label kind="guard" x="416" y="192">!split_at_work</label><label kind="synchronisation" x="136" y="192">c[out1]!</label><label kind="assignment" x="200" y="192">t[out1]=true,t[in]=false</label><nail x="192" y="216"/><nail x="128" y="216"/></transition></template>
'),
	println('<template><name x="5" y="5">andsplit</name><parameter>int in, int out1, int out2, int out3, int out4, int out5, int out6, int out7</parameter><location id="id11" x="296" y="240"></location><location id="id12" x="224" y="240"></location><location id="id13" x="152" y="240"></location><location id="id14" x="80" y="240"></location><location id="id15" x="8" y="240"></location><location id="id16" x="-56" y="240"></location><location id="id17" x="-120" y="240"></location><location id="id18" x="376" y="240"></location><init ref="id17"/><transition><source ref="id18"/><target ref="id17"/><label kind="guard" x="272" y="248">!split_at_work</label><label kind="synchronisation" x="-24" y="248">c[out7]!</label><label kind="assignment" x="64" y="248">t[out7]=true,t[in]=false</label><nail x="376" y="272"/><nail x="-120" y="272"/></transition><transition><source ref="id11"/><target ref="id18"/><label kind="synchronisation" x="312" y="216">c[out6]!</label><label kind="assignment" x="296" y="176">t[out6]=true</label></transition><transition><source ref="id12"/><target ref="id11"/><label kind="synchronisation" x="232" y="216">c[out5]!</label><label kind="assignment" x="224" y="192">t[out5]=true</label></transition><transition><source ref="id13"/><target ref="id12"/><label kind="synchronisation" x="160" y="216">c[out4]!</label><label kind="assignment" x="152" y="176">t[out4]=true</label></transition><transition><source ref="id14"/><target ref="id13"/><label kind="synchronisation" x="88" y="216">c[out3]!</label><label kind="assignment" x="80" y="200">t[out3]=true</label></transition><transition><source ref="id15"/><target ref="id14"/><label kind="synchronisation" x="16" y="216">c[out2]!</label><label kind="assignment" x="8" y="176">t[out2]=true</label></transition><transition><source ref="id16"/><target ref="id15"/><label kind="synchronisation" x="-48" y="216">c[out1]!</label><label kind="assignment" x="-56" y="192">t[out1]=true</label></transition><transition><source ref="id17"/><target ref="id16"/><label kind="synchronisation" x="-104" y="216">c[in]?</label></transition></template>'),
	println('<template><name>orsplit</name><parameter>int in, int out1, int out2, int out3, int out4, int out5, int out6, int out7</parameter><location id="id19" x="-408" y="-56"><urgent/></location><location id="id20" x="-472" y="-56"></location><location id="id21" x="-312" y="-56"></location><location id="id22" x="-8" y="-56"></location><init ref="id20"/><transition><source ref="id22"/><target ref="id20"/><label kind="synchronisation" x="-276" y="25">c[alwayswrite]!</label><label kind="assignment" x="-308" y="49">t[in]=false,split_at_work = false</label><nail x="-40" y="48"/><nail x="-472" y="48"/></transition><transition><source ref="id19"/><target ref="id21"/><label kind="guard" x="-408" y="-48">!split_at_work</label><label kind="assignment" x="-432" y="-32">split_at_work = true</label></transition><transition><source ref="id20"/><target ref="id19"/><label kind="synchronisation" x="-456" y="-80">c[in]?</label></transition><transition><source ref="id22"/><target ref="id22"/><label kind="synchronisation" x="-32" y="8">c[out7]!</label><label kind="assignment" x="-32" y="24">t[out7]=true</label><nail x="16" y="0"/><nail x="-8" y="0"/></transition><transition><source ref="id22"/><target ref="id22"/><label kind="synchronisation" x="64" y="8">c[out6]!</label><label kind="assignment" x="56" y="24">t[out6]=true</label><nail x="72" y="0"/><nail x="56" y="24"/></transition><transition><source ref="id22"/><target ref="id22"/><label kind="synchronisation" x="80" y="-24">c[out5]!</label><label kind="assignment" x="80" y="-8">t[out5]=true</label><nail x="80" y="-24"/><nail x="80" y="-8"/></transition><transition><source ref="id22"/><target ref="id22"/><label kind="synchronisation" x="80" y="-64">c[out4]!</label><label kind="assignment" x="80" y="-48">t[out4]=true</label><nail x="80" y="-56"/><nail x="80" y="-40"/></transition><transition><source ref="id22"/><target ref="id22"/><label kind="synchronisation" x="88" y="-96">c[out3]!</label><label kind="assignment" x="80" y="-80">t[out3]=true</label><nail x="80" y="-96"/><nail x="80" y="-72"/></transition><transition><source ref="id21"/><target ref="id22"/><label kind="synchronisation" x="-104" y="-160">c[out1]!</label><label kind="assignment" x="-224" y="-160">t[out1]=true</label><nail x="-312" y="-136"/><nail x="-32" y="-136"/></transition><transition><source ref="id21"/><target ref="id22"/><label kind="synchronisation" x="-104" y="-136">c[out2]!</label><label kind="assignment" x="-224" y="-136">t[out2]=true</label><nail x="-312" y="-112"/><nail x="-32" y="-112"/></transition><transition><source ref="id21"/><target ref="id22"/><label kind="synchronisation" x="-104" y="-112">c[out3]!</label><label kind="assignment" x="-224" y="-112">t[out3]=true</label><nail x="-312" y="-88"/><nail x="-40" y="-88"/></transition><transition><source ref="id21"/><target ref="id22"/><label kind="synchronisation" x="-104" y="-8">c[out7]!</label><label kind="assignment" x="-224" y="-8">t[out7]=true</label><nail x="-312" y="16"/><nail x="-32" y="16"/></transition><transition><source ref="id21"/><target ref="id22"/><label kind="synchronisation" x="-104" y="-32">c[out6]!</label><label kind="assignment" x="-224" y="-32">t[out6]=true</label><nail x="-312" y="-8"/><nail x="-40" y="-8"/></transition><transition><source ref="id21"/><target ref="id22"/><label kind="synchronisation" x="-104" y="-56">c[out5]!</label><label kind="assignment" x="-224" y="-56">t[out5]=true</label><nail x="-312" y="-32"/><nail x="-32" y="-32"/></transition><transition><source ref="id21"/><target ref="id22"/><label kind="synchronisation" x="-104" y="-80">c[out4]!</label><label kind="assignment" x="-224" y="-80">t[out4]=true</label></transition><transition><source ref="id22"/><target ref="id22"/><label kind="synchronisation" x="-16" y="-176">c[out1]!</label><label kind="assignment" x="-24" y="-160">t[out1]=true</label><nail x="-8" y="-136"/><nail x="32" y="-136"/></transition><transition><source ref="id22"/><target ref="id22"/><label kind="synchronisation" x="64" y="-136">c[out2]!</label><label kind="assignment" x="72" y="-120">t[out2]=true</label><nail x="72" y="-104"/><nail x="48" y="-128"/></transition></template>'),
	println('<template><name x="5" y="5">xorjoin</name><parameter>int in1, int in2, int in3, int in4, int in5, int in6, int in7, int out</parameter><location id="id23" x="-8" y="240"></location><location id="id24" x="304" y="240"></location><init ref="id23"/><transition><source ref="id24"/><target ref="id23"/><label kind="guard" x="192" y="152">!split_at_work</label><label kind="synchronisation" x="336" y="192">c[out]!</label><label kind="assignment" x="-16" y="136">t[out]=true,t[in1]=false,t[in2]=false,t[in3]=false,t[in4]=false,t[in5]=false,t[in6]=false,t[in7]=false</label><nail x="376" y="240"/><nail x="376" y="176"/><nail x="-8" y="176"/></transition><transition><source ref="id23"/><target ref="id24"/><label kind="synchronisation" x="112" y="272">c[in7]?</label><nail x="32" y="288"/><nail x="272" y="288"/></transition><transition><source ref="id23"/><target ref="id24"/><label kind="synchronisation" x="112" y="256">c[in6]?</label><nail x="32" y="272"/><nail x="272" y="272"/></transition><transition><source ref="id23"/><target ref="id24"/><label kind="synchronisation" x="112" y="240">c[in5]?</label><nail x="32" y="256"/><nail x="272" y="256"/></transition><transition><source ref="id23"/><target ref="id24"/><label kind="synchronisation" x="112" y="176">c[in1]?</label><nail x="24" y="192"/><nail x="280" y="192"/></transition><transition><source ref="id23"/><target ref="id24"/><label kind="synchronisation" x="112" y="192">c[in2]?</label><nail x="32" y="208"/><nail x="272" y="208"/></transition><transition><source ref="id23"/><target ref="id24"/><label kind="synchronisation" x="112" y="208">c[in3]?</label><nail x="24" y="224"/><nail x="264" y="224"/></transition><transition><source ref="id23"/><target ref="id24"/><label kind="synchronisation" x="112" y="224">c[in4]?</label></transition></template>
'),
	println('<template><name x="5" y="5">andjoin</name><parameter>int in1, int in2, int in3, int in4, int in5, int in6, int in7, int out</parameter><location id="id25" x="-16" y="240"></location><init ref="id25"/><transition><source ref="id25"/><target ref="id25"/><label kind="guard" x="48" y="256">!consumed[in7]</label><label kind="synchronisation" x="152" y="256">c[in7]?</label><label kind="assignment" x="200" y="256">consumed[in7]:=true</label><nail x="0" y="240"/><nail x="32" y="256"/><nail x="32" y="272"/></transition><transition><source ref="id25"/><target ref="id25"/><label kind="guard" x="48" y="224">!consumed[in6]</label><label kind="synchronisation" x="152" y="224">c[in6]?</label><label kind="assignment" x="200" y="224">consumed[in6]:=true</label><nail x="40" y="224"/><nail x="32" y="240"/></transition><transition><source ref="id25"/><target ref="id25"/><label kind="guard" x="48" y="192">!consumed[in5]</label><label kind="synchronisation" x="152" y="192">c[in5]?</label><label kind="assignment" x="200" y="192">consumed[in5]:=true</label><nail x="40" y="192"/><nail x="40" y="208"/></transition><transition><source ref="id25"/><target ref="id25"/><label kind="guard" x="48" y="152">!consumed[in4]</label><label kind="synchronisation" x="152" y="152">c[in4]?</label><label kind="assignment" x="200" y="152">consumed[in4]:=true</label><nail x="40" y="152"/><nail x="40" y="176"/></transition><transition><source ref="id25"/><target ref="id25"/><label kind="guard" x="48" y="112">!consumed[in3]</label><label kind="synchronisation" x="152" y="112">c[in3]?</label><label kind="assignment" x="200" y="112">consumed[in3]:=true</label><nail x="-8" y="232"/><nail x="24" y="112"/><nail x="32" y="136"/></transition><transition><source ref="id25"/><target ref="id25"/><label kind="guard" x="48" y="80">!consumed[in2]</label><label kind="synchronisation" x="152" y="80">c[in2]?</label><label kind="assignment" x="200" y="80">consumed[in2]:=true</label><nail x="8" y="80"/><nail x="16" y="104"/></transition><transition><source ref="id25"/><target ref="id25"/><label kind="guard" x="-96" y="304">!split_at_work &amp;&amp; consumed[in1] &amp;&amp; consumed[in2] &amp;&amp;
(in3 == noread || consumed[in3]) &amp;&amp; (in4 == noread || consumed[in4]) &amp;&amp;
(in5 == noread || consumed[in5]) &amp;&amp; (in6 == noread || consumed[in6]) &amp;&amp;
(in7 == noread || consumed[in7])</label><label kind="synchronisation" x="24" y="280">c[out]!</label><label kind="assignment" x="-104" y="368">consumed[in1]=false,consumed[in2]=false,consumed[in3]=false,consumed[in4]=false,consumed[in5]=false,consumed[in6]=false,consumed[in7]=false,
t[in1]=false,t[in2]=false,t[in3]=false,t[in4]=false,t[in5]=false,t[in6]=false,t[in7]=false,t[out]=true</label><nail x="-56" y="304"/><nail x="16" y="304"/></transition><transition><source ref="id25"/><target ref="id25"/><label kind="guard" x="48" y="48">!consumed[in1]</label><label kind="synchronisation" x="152" y="48">c[in1]?</label><label kind="assignment" x="200" y="48">consumed[in1]:=true</label><nail x="-16" y="56"/><nail x="0" y="56"/></transition></template>').


uppaal3c :- println('<template><name x="5" y="5">orjoin</name><parameter>int in1, int in2, int in3, int in4, int in5, int in6, int in7, int out</parameter><location id="id26" x="-16" y="240"></location><init ref="id26"/><transition><source ref="id26"/><target ref="id26"/><label kind="guard" x="48" y="256">!consumed[in7]</label><label kind="synchronisation" x="152" y="256">c[in7]?</label><label kind="assignment" x="200" y="256">consumed[in7]:=true</label><nail x="0" y="240"/><nail x="32" y="256"/><nail x="32" y="272"/></transition><transition><source ref="id26"/><target ref="id26"/><label kind="guard" x="48" y="224">!consumed[in6]</label><label kind="synchronisation" x="152" y="224">c[in6]?</label><label kind="assignment" x="200" y="224">consumed[in6]:=true</label><nail x="40" y="224"/><nail x="32" y="240"/></transition><transition><source ref="id26"/><target ref="id26"/><label kind="guard" x="48" y="192">!consumed[in5]</label><label kind="synchronisation" x="152" y="192">c[in5]?</label><label kind="assignment" x="200" y="192">consumed[in5]:=true</label><nail x="40" y="192"/><nail x="40" y="208"/></transition><transition><source ref="id26"/><target ref="id26"/><label kind="guard" x="48" y="152">!consumed[in4]</label><label kind="synchronisation" x="152" y="152">c[in4]?</label><label kind="assignment" x="200" y="152">consumed[in4]:=true</label><nail x="40" y="152"/><nail x="40" y="176"/></transition><transition><source ref="id26"/><target ref="id26"/><label kind="guard" x="48" y="112">!consumed[in3]</label><label kind="synchronisation" x="152" y="112">c[in3]?</label><label kind="assignment" x="200" y="112">consumed[in3]:=true</label><nail x="-8" y="232"/><nail x="24" y="112"/><nail x="32" y="136"/></transition><transition><source ref="id26"/><target ref="id26"/><label kind="guard" x="48" y="80">!consumed[in2]</label><label kind="synchronisation" x="152" y="80">c[in2]?</label><label kind="assignment" x="200" y="80">consumed[in2]:=true</label><nail x="8" y="80"/><nail x="16" y="104"/></transition><transition><source ref="id26"/><target ref="id26"/><label kind="guard" x="-96" y="304">!split_at_work &amp;&amp; (consumed[in1]||consumed[in2]||consumed[in3]||consumed[in4]||consumed[in5]||consumed[in6]||consumed[in7]) &amp;&amp;canfire(in1,in2,in3,in4,in5,in6,in7,out)</label><label kind="synchronisation" x="24" y="280">c[out]!</label><label kind="assignment" x="-96" y="352">consumed[in1]=false,consumed[in2]=false,consumed[in3]=false,consumed[in4]=false,consumed[in5]=false,consumed[in6]=false,consumed[in7]=false,t[in1]=false,t[in2]=false,t[in3]=false,t[in4]=false,t[in5]=false,t[in6]=false,t[in7]=false,t[out]=true</label><nail x="-56" y="304"/><nail x="16" y="304"/></transition><transition><source ref="id26"/><target ref="id26"/><label kind="guard" x="48" y="48">!consumed[in1]</label><label kind="synchronisation" x="152" y="48">c[in1]?</label><label kind="assignment" x="200" y="48">consumed[in1]:=true</label><nail x="-16" y="56"/><nail x="0" y="56"/></transition></template>
').


uppaal3d :- println('<template><name>Dummies</name><location id="id27" x="0" y="0"></location><init ref="id27"/><transition><source ref="id27"/><target ref="id27"/><label kind="synchronisation" x="-104" y="0">c[alwayswrite]?</label><nail x="-88" y="0"/><nail x="-88" y="-40"/></transition><transition><source ref="id27"/><target ref="id27"/><label kind="synchronisation" x="24" y="-40">c[alwaysread]!</label><nail x="0" y="-32"/><nail x="96" y="0"/></transition></template>
'),
	println('<template><name>Order</name><parameter>int in1, int in2</parameter><location id="id27" x="-24" y="-80"></location><location id="id28" x="80" y="-80"><name x="70" y="-110">second</name></location><location id="id29" x="-136" y="-80"><name x="-146" y="-110">first</name></location><location id="id30" x="-232" y="-80"></location><init ref="id30"/><transition><source ref="id27"/><target ref="id28"/><label kind="guard" x="8" y="-104">t[in2]</label></transition><transition><source ref="id29"/><target ref="id27"/><label kind="guard" x="-112" y="-104">not t[in1]</label></transition><transition><source ref="id30"/><target ref="id29"/><label kind="guard" x="-200" y="-104">t[in1]</label></transition></template>'),
	println('<system>').

uppaal4 :- 
	(remove_events;
	(findall(X,uppaal_startevent(X),_),nl,
	findall(X,uppaal_endevent(X),_),nl,
	findall(X,uppaal_activity(X),_),nl,
	findall(X,uppaal_split(X),_),nl,
	findall(X,uppaal_join(X),_),
	nl,
	uppaal_system_def)).

uppaal5 :-println('</system></nta>').



channels_for_arcs_to_orjoins(ArcsToOrjoins) :-
	findall([X,Y],
	(
		nth0(N,ArcsToOrjoins,[X,Y]) ,
		print('const int '),
		print(X),print('_'),print(Y),
		print(' = '),print(N),print(';'),
		nl
	)
	,_).

channels_for_arcs_to_local_elements(ArcsToLocalElements,NumberOfArcsToOrjoins) :-
	findall([X,Y],
	(
		nth0(N,ArcsToLocalElements,[X,Y]) ,
		print('const int '),
		print(X),print('_'),print(Y),
		print(' = '),NN is (N+NumberOfArcsToOrjoins),print(NN),print(';'),
		nl
	)
	,_).



print_uppaal_structure(ArcsToOrjoins,ArcsToLocalElements) :-
	append(ArcsToOrjoins,ArcsToLocalElements,AllArcs),
	findall([X,Y],
	(
		member([X,Y],ArcsToOrjoins),
		print('{'),
		findall([From,To],
		(
			member([From,To],AllArcs),
			bpmn2semantics(X,Y,To),
			(not(last(AllArcs,[From,To])),print(','))
		),
		_),
		print('}'),(not(last(ArcsToOrjoins,[X,Y])),print(',')),nl
	),
	_).

bpmn2semantics(X,Y,To) :- % [X,Y] - input arc from OR-join, [From,To] - arc that might carry a token
	((secondpath(To,X,_,[Y]),To \==Y )  -> print(1);
	To==X              -> print(1);
	not(secondpath(To,X,_,[Y])) -> print(0);
	(secondpath(To,X,_,[Y]),To ==Y  ) -> print(0)
	).
		






% some useful facts about arcs
uarc(X,Y) :- arc(X,Y);arc(Y,X).
% more_than_one_incoming_arcs(X) :- arc(A,X),arc(B,X),A\==B.
% more_than_one_outgoing_arcs(X) :- arc(X,A),arc(X,B),A\==B.
more_than_one_incoming_arcs(X) :- findall(A,arc(A,X),L),L = [_|L2], L2 \== [].
more_than_one_outgoing_arcs(X) :-  findall(A,arc(X,A),L),L = [_|L2], L2 \== [].
no_incoming_arcs(X) :- not(arc(_,X)).
no_outgoing_arcs(X) :- not(arc(X,_)).
% successor(X,Y) means that X is followed by Y (possibly via some connectors)
% (Note: property9 must be tested before calling successor(X,Y) in order to avoid infinite cycles.)
successor(X,Y) :- arc(X,Y).
successor(X,Y) :- arc(X,C),connector(C),successor(C,Y).
% types of elements
orconnector(X)  :- clause(or(X),true).
andconnector(X) :- clause(and(X),true).
xorconnector(X) :- clause(xor(X),true).
connector(X)    :- clause(and(X),true) ; clause(or(X),true) ; clause(xor(X),true).
split(X) :- connector(X),more_than_one_outgoing_arcs(X). % When using this, we assume that property8 already has been tested.
join(X) :- connector(X),more_than_one_incoming_arcs(X).  % When using this, we assume that property8 already has been tested.
orsplit(X)      :- orconnector(X),split(X).
andsplit(X)     :- andconnector(X),split(X).
xorsplit(X)     :- xorconnector(X),split(X).
orjoin(X)       :- orconnector(X),join(X).
andjoin(X)      :- andconnector(X),join(X).
xorjoin(X)      :- xorconnector(X),join(X).

type(X,Type)    :- (orconnector(X),Type=or);(andconnector(X),Type=and);(xorconnector(X),Type=xor).
have_same_type(X,Y) :- type(X,TypeX),type(Y,TypeY),TypeX=TypeY.
element(X) :- event(X);function(X);connector(X).

startevent(X) :- event(X),no_incoming_arcs(X).
endevent(X) :- event(X),no_outgoing_arcs(X).

binarysplit(X) :- split(X),findall(Y,arc(X,Y),L),length(L,2).
binaryjoin(X) :- join(X),findall(Y,arc(Y,X),L),length(L,2).

number_of_arcs(N) :-  findall([A,B],arc(A,B),L),length(L,N).

after(Set,Stop,List) :- descend(Set,[],Stop),List = Set.
after(Set,Stop,List) :- descend(Set,New,Stop),New\==[],ord_union(Set,New,Setplus),after(Setplus,Stop,List).

descend(Set,New,Stop) :- findall(X,(member(A,Set),arc(A,X),X\==Stop,not(ord_memberchk(X,Set))),New0),
		list_to_ord_set(New0,New).


before(Set,Stop,List) :- ascend(Set,[],Stop),List = Set.
before(Set,Stop,List) :- ascend(Set,New,Stop),New\==[],ord_union(Set,New,Setplus),before(Setplus,Stop,List).


ascend(Set,New,Stop) :- findall(X,(member(A,Set),arc(X,A),X\==Stop,not(ord_memberchk(X,Set))),New0),
		list_to_ord_set(New0,New).


sese(S,J,After2) :- path(S,J,_),!,after([S],J,After),before([J],S,Before),
		ord_del_element(After,S,After2),ord_del_element(Before,J,Before2),After2==Before2.


% paths and reachability
path(A,B,Path) :-
       travel(A,B,[A],Q), 
       reverse(Q,Path).

travel(A,B,P,[B|P]) :- 
       arc(A,B).
travel(A,B,Visited,Path) :-
       arc(A,C),
       C \== B,
       \+member(C,Visited),
       travel(C,B,[C|Visited],Path).

secondpath(A,B,Path,Exclude) :-
       travel2(A,B,[A],Q,Exclude), 
       reverse(Q,Path).

travel2(A,B,P,[B|P],_) :- 
       arc(A,B).
travel2(A,B,Visited,Path,Exclude) :-
       arc(A,C),not(ord_memberchk(C,Exclude)),
       C \== B,
       \+member(C,Visited),
       travel2(C,B,[C|Visited],Path,Exclude).




%-----------------------------------------------------------------------------------------------------------
% REDUKTIONSREGELN
%-----------------------------------------------------------------------------------------------------------


:- dynamic region/1.             % per Graphreduktion als korrekt erkannter Modellteil
% Es kann sinnvoll werden, Regions zu cachen.
% Reduktionsregeln f�r Sequenzen:
% fasst Sequenzen von Events und Funktionen zu einer einzigen Region zusammen:

region(X) :- event(X).
region(X) :- function(X).

forget(Me) :- %  print('Vergiss: '),print(Me),nl,
	((arc(In,Me),retract(arc(In,Me)),fail);true), % Alle Arcs von und zu diesem Element vergessen
	((arc(Me,Out),retract(arc(Me,Out)),fail);true),
	(retract(region(Me));retract(function(Me));retract(event(Me));
	 retract(xor(Me));retract(or(Me));retract(and(Me))).


reduce_sequence :- region(M1), arc(M1,M2), region(M2),
           arc(In,M1),arc(M2,Out),                                               % findet folglich keine Start- und Endereignisse
           atom_concat(M1,M2,S),assert(region(S)),
           print('Reduce Sequence: '),print(S),nl,
           assert(arc(In,S)),
           assert(arc(S,Out)),
           retract(arc(M1,M2)),
	forget(M1),forget(M2).


% Reduktionsregel für korrekte Blöcke

split_join_block(Split,Join,Split_out) :-
		  findall(X1,arc(Split,X1),Li1), delete(Li1,Join,Li1_),sort(Li1_,Split_out), % Das delete ist notwenig, da ein Pfeil direkt vom Split zum Join zeigen kann.
		  not(list_contains_connectors(Split_out)), % Dann k�nnen nur Regions folgen.
                  findall(X2,arc(X2,Join),Li2),delete(Li2,Split,Li2_),sort(Li2_,Join_in),
                  Split_out == Join_in.

list_contains_connectors(List) :- (member(C,List),connector(C)).

reduce_wellstructured_split_join_block :-   ((xorsplit(Split),xorjoin(Join));
		 (orsplit(Split),orjoin(Join));
		 (andsplit(Split),andjoin(Join))),
		split_join_block(Split,Join,Regions),
           	arc(In,Split),arc(Join,Out),
           	type(Split,Type),
		concat_atom([begin_,Type,Split,Join,end_,Type],S),
          	assert(region(S)),print('Reduce '),print(Type),print('-block: '),print(S),nl,
           	assert(arc(In,S)),
	   	assert(arc(S,Out)),
		forget(Split),forget(Join),
                ((member(Member,Regions),forget(Member),fail);true).

reduce_correct_split_join_block :-  orjoin(Join),(andsplit(Split);xorsplit(Split)),
		split_join_block(Split,Join,Regions),
           	arc(In,Split),arc(Join,Out),
           	type(Split,Type),
		assertz(unstructured),
		print('[H] Consider to replace '),print(Join),print(' by an '),
		print(Type),print('-join  (WARNING)'),nl,
		concat_atom([begin_,Type,Split,Join,end_,Type],S),
          	assert(region(S)),print('Reduce '),print(Type),print('->OR-block: '),print(S),nl,
           	assert(arc(In,S)),
	   	assert(arc(S,Out)),
		forget(Split),forget(Join),
                ((member(Member,Regions),forget(Member),fail);true).

% Mindestens drei verschiedene Vorkommen Split -> Region -> Join (unabh. vom Typ der Splits und Joins)
reduce_big_split_join_block :- split(Split),
		arc(Split,R1),region(R1),arc(Split,R2),region(R2),R2 \== R1, arc(Split,R3), region(R3), R3 \== R1,R3 \== R2,
		arc(R1,Join),join(Join),arc(R2,Join),arc(R3,Join),
		forget(R1),
		print('Between split '),print(Split),print(' and join '),print(Join),
		print(' are more than two arcs split->node->join. We delete one of them.'),nl.
		

% Reduktionsregel f�r Iterationen (nur f�r Iterationen, bei denen zwischen XOR-Split und -Join
% h�chstens ein Event, eine Funktion oder eine Region liegt

reduce_wellstructured_iteration :- xorjoin(X1),xorsplit(X2),binaryjoin(X1),binarysplit(X2),
		(arc(X1,X2);(arc(X1,R1),arc(R1,X2),not(connector(R1)))),
		(arc(X2,X1);(arc(X2,R2),arc(R2,X1),not(connector(R2)))),
		arc(In,X1),In \== X2, In \== R2, arc(X2,Out), Out \== X1, Out \== R2,
		concat_atom([begin_iteration_,X1,X2,end_iteration_],S),
          	assert(region(S)),print('Reduce iteration block: '),print(S),nl,
		forget(X1),forget(X2),
		((nonvar(R1),forget(R1));true),
		((nonvar(R2),forget(R2));true),
		assert(arc(In,S)),assert(arc(S,Out)).

% Seiteneffekt: assertz(unstructured)

reduce_correct_iteration :- orjoin(X1),xorsplit(X2),binaryjoin(X1),binarysplit(X2),
		(arc(X1,X2);(arc(X1,R1),arc(R1,X2),not(connector(R1)))),
		(arc(X2,X1);(arc(X2,R2),arc(R2,X1),not(connector(R2)))),
		arc(In,X1),In \== X2, In \== R2, arc(X2,Out), Out \== X1, Out \== R2,
		concat_atom([begin_iteration_,X1,X2,end_iteration_],S),
		print('[N] consider to replace the loop entry '),print(X2),print(' by an XOR-join'),nl,
		assertz(unstructured),
          	assert(region(S)),print('Reduce iteration block: '),print(S),nl,
		forget(X1),forget(X2),
		((nonvar(R1),forget(R1));true),
		((nonvar(R2),forget(R2));true),
		assert(arc(In,S)),assert(arc(S,Out)).		


merge_startevent_region_join :- startevent(A),arc(A,R),arc(R,Join),join(Join),region(R),
				atom_concat(A,R,AR),
				forget(A),forget(R),assert(event(AR)),assert(arc(AR,Join)),
				print('Merging start event '),print(A),print(' with region '),
				print(R),nl.

           
merge_startevents :- startevent(A),arc(A,Join),join(Join),arc(B,Join),A\==B,startevent(B),!,
                                type(Join,Type),
                                print('Merge start-events '),print(A),print(' and '),print(B),nl,
				concat_atom([merged_startevent_,Type,A,B],S),
                                assert(event(S)),
                                forget(A),forget(B),
                                % Wenn am Join nix weiter ankommt, kann er verschwinden:
				(
				 (not(arc(_,Join)),arc(Join,After),forget(Join),assert(arc(S,After)))
				;
				 assert(arc(S,Join))
				).

merge_split_region_endevent :- endevent(A),arc(R,A),arc(Split,R),split(Split),region(R),
				atom_concat(R,A,RA),
				forget(A),forget(R),assert(event(RA)),assert(arc(Split,RA)),
				print('Merging region '),print(R),print(' with end event '),
				print(A),nl.

merge_endevents :- endevent(A),arc(Split,A),split(Split),arc(Split,B),A\==B,endevent(B),!,
                                type(Split,Type),
                                print('Merge end-events '),print(A),print(' and '),print(B),nl,
				concat_atom([merged_endevent_,Type,A,B],S),
                                assert(event(S)),
                                forget(A),forget(B),
				(
				 (not(arc(Split,_)),arc(Before,Split),forget(Split),assert(arc(Before,S)))
				;
				 assert(arc(Split,S))
				).

delete_and_endevent :- andsplit(S),arc(S,End),endevent(End),not(path(S,S,_)),
		arc(In,S),arc(S,Out),Out \== End,
		print('Delete endevent '),print(End),print(' after AND-split '),print(S),nl,
		((binarysplit(S),forget(S),assert(arc(In,Out)));true), %Binary Splits vergessen, sonst nur das Endevent vergessen
		forget(End).


reduce_homogeneous :- split(S),join(J),sese(S,J,Between),have_same_type(S,J),
		% In Between d�rfen dann nur Konnektoren des selben Typs liegen:
		not((connector(C),not(have_same_type(C,S)),ord_memberchk(C,Between))),
		% Ist der gemeinsame Typ AND- oder OR, sind au�erdem Zyklen verboten:
		(xorsplit(S)
		 ; 
		(not(( ord_memberchk(Join,Between),join(Join),path(Join,Join,_)  )))
		),
		arc(In,S),arc(J,Out),
		print('Reduce homogeneous block between split '),print(S),print(' and join '),print(J),
		nl,
		forget(S),forget(J),
		((member(Member,Between),forget(Member),fail);true), %ord_memberchk does not work here!
		assert(arc(In,Out)).
		


% Wende Reduktionsregeln wiederholt an.
% Wohlgeformte Modelle werden reduziert zu: Startevent -> Region -> Endevent
% Reduktionsregeln liefern true, wenn Reduzierung stattfand, sonst false

re :- (
          (reduce_sequence,!);
	  (reduce_wellstructured_split_join_block,!);
          (reduce_correct_split_join_block,!);
          (merge_startevent_region_join,!);
	  (merge_split_region_endevent,!);
          (merge_startevents,!);
          (merge_endevents,!);
	  (reduce_wellstructured_iteration,!);
	  (reduce_correct_iteration,!);
	  (reduce_big_split_join_block,!);
	  (delete_and_endevent,!);
	  (reduce_homogeneous,!)
       ),
        re;reduced.



re_yawl :- (
          (reduce_sequence,!);
	  (reduce_wellstructured_split_join_block,!);
          (merge_startevent_region_join,!);
	  (merge_split_region_endevent,!);
          (merge_startevents,!);
          (merge_endevents,!);
	  (reduce_wellstructured_iteration,!);
	  (reduce_big_split_join_block,!);
	  (delete_and_endevent,!);
	  (reduce_homogeneous,!)
       ),
        re_yawl.

reduced :-  findall([A,B],arc(A,B),L),not(nth1(3,L,_)),              % Es gibt nur 2 Arcs
                    startevent(S),region(R),endevent(E),arc(S,R),arc(R,E).

% Remove all events (before converting the model into the input language of a model checker):
remove_events :- event(E),arc(In,E),arc(E,Out),	% event is neither start event nor end event
		forget(E),
		((not(arc(In,Out)),assert(arc(In,Out)));true),
		remove_events.

% Remove orphean connectors (that might exist after removing events inside a control block) as well:
remove_events :- connector(X),not(split(X)),not(join(X)),
		arc(In,X),arc(X,Out),
		forget(X),
		((not(arc(In,Out)),assert(arc(In,Out)));true),
		remove_events.

% Printing some useful information
v(Element) :- event(Element),print('Event '),print(Element),print(': '),elementname(Element,Name),print(Name).
v(Element) :- function(Element),print('Function '),print(Element),print(': '),elementname(Element,Name),print(Name).
v(Element) :- split(Element),type(Element,Type),print('========'),print(Type),print('-Split '),print(Element),nl,nachbarn(Element),print('========'),print(Type),print('-Split '),print(Element).
v(Element) :- join(Element),type(Element,Type),print('========'),print(Type),print('-Join '),print(Element),nl,nachbarn(Element),print('========'),print(Type),print('-Join '),print(Element).

nachbarn(Element) :- print('--------davor:'),nl,
			findall(In,(arc(In,Element),print(' *'),vprint(In),print(','),nl),_),nl,
			print('--------danach: '),nl,
			findall(Out,(arc(Element,Out),print(' *'),vprint(Out),print(','),nl),_).

vprint(Element) :- (event(Element);function(Element)),v(Element).
vprint(Element) :- split(Element),type(Element,Type),print(Type),print('Split '),print(Element).
vprint(Element) :- join(Element),type(Element,Type),print(Type),print('Join '),print(Element).